/*
Copyright 1989 by the Movies of the Future Group
of the MIT Media Laboratory
All rights reserved.

Permission to use, copy, and modify this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of M.I.T. not be used in
advertising or publicity pertaining to distribution of the software
without specific, written prior permission.  This software or any
other copies thereof may not be provided or otherwise made available to
others.  No title to or ownership of the software is hereby transferred.
M.I.T. makes no representations about the suitability of this software for any
purpose.  It is provided "as is" without express or implied warranty.
*/
/* dxopen - extended datfile open

DATFILE *dxopen(fname, mode)
char name[], mode[];

The programming needed to set up subarrays can be tedious.  dxopen
automatically parses "filename[range_spec]" and generates calls to
d_open and d_submatrix.  If dxopen is substituted for d_open, d_read for
read, and d_getsubdim for d_getdim, subarray/subsampled I/O becomes
virtually transparent to the program.

fname is a datfile name possibly suffixed by a bracketed list
of ranges separated by blank or comma or any combination:
	[range range ...]
or	[range,range,...]
or	[range][range]...
The syntax of range is defined in getrange(). If ranges are not given, dxopen
behaves just like d_open.  If ranges are given, then fname must already
exist and will not be deleted even if mode "w" is selected; the appropriate
submatrix will be set up from the range_spec.

Limitations: no provision for DELETEDIM.

If compiled with -DNORETURN, does exit(-1) instead of return(NULL)
upon error.
*/

#include "dat.h"
#include <ctype.h>
#define START '['		/* start of range list */
#define STOP ']'		/* end of range list (optional) */
#define BLANK ' '		/* range separator */
#define SEP ','			/* alternate range separator */

DATFILE *dxopen(name, mode)
const char *name;
const char *mode;
{
	char fname[128], buf[16];
	const register char *p;
	register char *b;
	DATFILE *df;
	
	p = name;

	for (b=fname; *p!='\0' && *p!=START; )	/* extract fname */
		*b++ = *p++;
	*b = '\0';
	if (*p!='\0' && mode[0]=='w')	/* if range given, change "w" to "a" */
		mode = mode[1]=='+' ? "a+" : "a";
	if ((df = d_open(fname, mode)) == (DATFILE *)NULL)
#ifdef NORETURN
		exit(-1);
#else
		return(NULL);
#endif

	if (*p!='\0') {			/* if a range is given, set up submatrix */
		struct d_addr start, stop, step;
		int ndims, i;
		char flag;
		if ((ndims = d_getsubdim(df, &stop)) <= 0) {
			fprintf(stderr,"%s: no subarray allowed on new file\n", fname);
#ifdef NORETURN
			exit(-1);
#else
			return(NULL);
#endif
		}
		start.dset = start.chan = 0;
		step.dset = step.chan = 1;
		for (i=0; i<ndims; i++) {
			start.dim[i] = 0;
			step.dim[i] = 1;
		}
		p++;			/* skip opening delimiter */
		for (i=0; *p!='\0'; ) {
			while (*p==BLANK)		/* skip blanks, if any */
				p++;
			switch (flag = *p) {
			case 'd': case 'c':		/* "d[ataset]" or "c[hannel]" */
				while (isalpha(*p))	/* skip word	*/
					p++;
				while (*p==BLANK)	/* skip blanks, if any */
					p++;
				break;
			}
				
									/* copy a range specifier to buf[] */
			for (b=buf; *p!='\0' && *p!=BLANK && *p!=SEP && *p!=STOP; )
 				*b++ = *p++;
			*b = '\0';
			while (*p==BLANK)	/* skip blanks, if any */
				p++;
			if (*p==SEP)		/* skip delimiters */
				p++;
			else if (*p==STOP && *++p==START)
				p++;
					
			switch (flag) {			/* with thanks to J. Wang */
			case 'd':
				getrange(fname,buf,
					&start.dset,&stop.dset,stop.dset,&step.dset);
				break;
			case 'c':
				getrange(fname,buf,
					&start.chan,&stop.chan,stop.chan,&step.chan);
				break;
			default:
				if (i >= ndims)
					fprintf(stderr,"%s: too many dimensions specified\n",fname);
				else {
				getrange(fname,buf,
					&start.dim[i],&stop.dim[i],stop.dim[i],&step.dim[i]);
				i++;
				}
				break;
			}
		}
#if DEBUG & 1
    fprintf(stderr,"start\tstop\tstep\n");
    fprintf(stderr,"%D\t%D\t%D\n", start.dset, stop.dset, step.dset);
    fprintf(stderr,"%D\t%D\t%D\n", start.chan, stop.chan, step.chan);
    for (i = 0; i < ndims; i++) 
	fprintf(stderr,"%D\t%D\t%D\n", start.dim[i], stop.dim[i], step.dim[i]);
#endif
		d_submatrix(df,&start,&stop,&step);
	}
	return(df);
}
